__all__ = ()

from ..guild.constants import DESCRIPTION_LENGTH_MAX, NAME_LENGTH_MIN, NAME_LENGTH_MAX
from ..guild_badge.constants import (
    BADGE_TAG_LENGTH_MIN as BADGE_TAG_LENGTH_MIN, BADGE_TAG_LENGTH_MAX as BADGE_TAG_LENGTH_MAX
)
