__all__ = ()

APPLICATION_ROLE_CONNECTION_KEY = 'guild_connections'
BOOSTER_KEY = 'premium_subscriber'
BOT_ID_KEY = 'bot_id'
INTEGRATION_ID_KEY = 'integration_id'
PURCHASABLE_KEY = 'available_for_purchase'
SUBSCRIPTION_LISTING_ID_KEY = 'subscription_listing_id'
