__all__ = ()

from .preinstanced import ButtonStyle, SeparatorSpacingSize, TextInputStyle


BUTTON_STYLE_DEFAULT = ButtonStyle.blue
SEPARATOR_SPACING_SIZE_DEFAULT = SeparatorSpacingSize.small
TEXT_INPUT_STYLE_DEFAULT = TextInputStyle.short

DESCRIPTION_LENGTH_MIN = 0
DESCRIPTION_LENGTH_MAX = 1024

LABEL_LENGTH_MAX__BUTTON = 80
LABEL_LENGTH_MAX__LABEL = 45

PLACEHOLDER_LENGTH_MAX = 150

OPTIONS_LENGTH_MIN = 1
OPTIONS_LENGTH_MAX = 25

MAX_LENGTH_MIN = 0
MAX_LENGTH_MAX = 4000
MAX_LENGTH_DEFAULT = 0

MIN_LENGTH_MIN = 0
MIN_LENGTH_MAX = 4000
MIN_LENGTH_DEFAULT = 0

MAX_VALUES_MIN = 1
MAX_VALUES_MAX__ATTACHMENT_INPUT = 10
MAX_VALUES_MAX__SELECT = 25
MAX_VALUES_DEFAULT = 1

MIN_VALUES_MIN = 0
MIN_VALUES_MAX__ATTACHMENT_INPUT = 15
MIN_VALUES_MAX__SELECT = 25
MIN_VALUES_DEFAULT = 1

VALUE_LENGTH_MAX = 4000

URL_LENGTH_MAX = 512

CONTENT_LENGTH_MIN = 1
CONTENT_LENGTH_MAX = 4000
