__all__ = ()

from .compatibility import AUTO_RELOAD_SUPPORTED, INotifyFlags


AUTO_RELOAD_DELAY = 5.0


if AUTO_RELOAD_SUPPORTED:
    WATCH_MASK_CREATE = INotifyFlags.CREATE
    WATCH_MASK_DELETE = INotifyFlags.DELETE
    WATCH_MASK_DELETE_SELF = INotifyFlags.DELETE_SELF
    WATCH_MASK_MOVE_FROM = INotifyFlags.MOVED_FROM
    WATCH_MASK_MOVE_SELF = INotifyFlags.MOVE_SELF
    WATCH_MASK_MOVE_TO = INotifyFlags.MOVED_TO
    WATCH_MASK_UPDATE = INotifyFlags.MODIFY
    
else:
    WATCH_MASK_CREATE = 0
    WATCH_MASK_DELETE = 0
    WATCH_MASK_DELETE_SELF = 0
    WATCH_MASK_MOVE_FROM = 0
    WATCH_MASK_MOVE_SELF = 0
    WATCH_MASK_MOVE_TO = 0
    WATCH_MASK_UPDATE = 0


WATCH_MASK_GENERAL = (
    WATCH_MASK_CREATE | WATCH_MASK_DELETE | WATCH_MASK_DELETE_SELF | WATCH_MASK_MOVE_FROM | WATCH_MASK_MOVE_SELF |
    WATCH_MASK_MOVE_TO | WATCH_MASK_UPDATE
)
