__all__ = ()

AUTO_ARCHIVE_DEFAULT = 3600
AUTO_ARCHIVE_OPTIONS = frozenset((3600, 86400, 259200, 604800))

BITRATE_MIN = 8000
BITRATE_MAX = 384000
BITRATE_DEFAULT = 48000

NAME_LENGTH_MIN = 1
NAME_LENGTH_MAX = 100
NAME_DEFAULT = 'channel'

SLOWMODE_MIN = 0
SLOWMODE_MAX = 21600
SLOWMODE_DEFAULT = 0

STATUS_LENGTH_MIN = 0
STATUS_LENGTH_MAX = 500

TOPIC_LENGTH_MIN = 0
TOPIC_LENGTH_MAX = 1024

USER_LIMIT_MIN = 0
USER_LIMIT_MAX = 99
USER_LIMIT_DEFAULT = 0
