__all__ = ()

DISCRIMINATOR_VALUE_MIN = 0
DISCRIMINATOR_VALUE_MAX = 9999

DISPLAY_NAME_LENGTH_MAX = 32

NAME_LENGTH_MIN = 2
NAME_LENGTH_MAX = 32

WEBHOOK_NAME_LENGTH_MIN = 1
WEBHOOK_NAME_LENGTH_MAX = 80
